from typing import List
from argparse import ArgumentParser
import csv
import os

import utils.table_utils
from utils import table_utils
from utils.table_utils import get_column
from trajectory.trajectory import Trajectory
from BioCycle import calculate_offsets

import numpy as np


def main():
    # parse arguments
    parser: ArgumentParser = ArgumentParser()
    parser.add_argument('--input_file', type=str, required=True, help="")
    parser.add_argument('--output_dir', type=str, required=True, help="")
    parser.add_argument('--methods', type=str, required=True, help="")

    parser.add_argument('--start_period', type=float, default=20.0, help="The lower bound of the range of periods to "
                                                                         "search for.")
    parser.add_argument('--end_period', type=float, default=28.0, help="The upper bound of the range of periods to "
                                                                       "search for")

    args = parser.parse_args()

    # TODO start and end periods
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)

    # parse file
    header, rows = table_utils.open_table(args.input_file)

    # get trajectories
    tps, tp_reps, num_reps, _ = utils.table_utils.get_timepoints(args.input_file)
    trajs: List[Trajectory] = table_utils.get_trajs(args.input_file)

    # write input data file
    method_file_postfix: str = 'input_data'
    input_fname: str = "%s/%s.csv" % (args.output_dir, method_file_postfix)

    tp_reps_str_l: List[str] = [str(x) for x in tp_reps]

    header = [header[0]] + tp_reps_str_l
    rows: List[List[str]] = table_utils.trajs_to_rows(trajs)
    table_utils.write_table(input_fname, header, rows, delimiter=",")

    # run MetaCycle
    methods: List[str] = args.methods.split(",")
    methods = [method.upper() for method in methods]
    tp_reps_str: str = ",".join(tp_reps_str_l)

    cmd: str = "Rscript MetaCycle/MetaCycle.R --input_file %s --output_dir %s " \
               "--timepoints_reps %s --methods %s" % (input_fname, args.output_dir, tp_reps_str, ",".join(methods))
    print(cmd)
    os.system(cmd)

    # read MetaCycle output
    mc_file_name_csv: str = "%s/meta2d_%s.csv" % (args.output_dir, method_file_postfix)
    with open(mc_file_name_csv) as file:
        reader = csv.reader(file, delimiter=',')
        rows_mc: List[List[str]] = [row for row in reader]
    header_mc: List[str] = rows_mc.pop(0)

    mc_file_name: str = "%s/meta2d_%s.tsv" % (args.output_dir, method_file_postfix)
    table_utils.write_table(mc_file_name, header_mc, rows_mc)

    if len(methods) == 1:
        method_read: str = methods[0]
    else:
        method_read: str = "meta2d"

    pvals: List[str] = get_column(mc_file_name, "%s_pvalue" % method_read)
    qvals: List[str] = get_column(mc_file_name, "%s_qvalue" % method_read)
    periods: List[str] = get_column(mc_file_name, "%s_period" % method_read)

    if method_read == "meta2d":
        phases: List[str] = get_column(mc_file_name, "%s_phase" % method_read)
        amplitudes: List[str] = get_column(mc_file_name, "%s_AMP" % method_read)
    else:
        phases: List[str] = get_column(mc_file_name, "%s_adjphase" % method_read)
        amplitudes: List[str] = get_column(mc_file_name, "%s_amplitude" % method_read)

    phases = ['0' if np.isnan(float(x)) else x for x in phases]
    amplitudes = ['0' if np.isnan(float(x)) else x for x in amplitudes]

    offsets: List[str] = []
    for idx in range(len(trajs)):
        try:
            traj = trajs[idx]
            traj.compute_statistics()
            traj.normalize()

            periods_np = np.array([float(periods[idx])])
            phases_np = np.array([float(phases[idx])])
            amplitudes_np = np.array([float(amplitudes[idx])])

            offset = calculate_offsets([trajs[idx]], periods_np, phases_np, amplitudes_np)[0]
            offsets.append(offset)
        except ValueError:
            offsets.append("NA")

    # write output file
    output_fname: str = "%s/results.tsv" % args.output_dir
    print("Writing output to %s" % output_fname)

    header_out = [header[0]] + ["P_VALUE", "Q_VALUE", "PERIOD", "PHASE", "AMPLITUDE", "OFFSET"]
    rows_out: List[List[str]] = []
    for idx in range(len(pvals)):
        row: List[str] = [rows[idx][0]] + [pvals[idx], qvals[idx], periods[idx], phases[idx], amplitudes[idx],
                                           offsets[idx]]
        rows_out.append(row)

    table_utils.write_table(output_fname, header_out, rows_out)

    """
    # remove extra files
    os.remove(input_fname)
    os.remove(mc_file_name)
    for method in methods:
        method_file_name: str = "%s/%sresult_%s.csv" % (args.output_dir, method.upper(), method_file_postfix)
        os.remove(method_file_name)
    """


if __name__ == "__main__":
    main()
